package com.byagowi.persiancalendar.ui.common

import android.content.ClipData
import android.content.Context
import androidx.compose.animation.AnimatedContent
import androidx.compose.animation.AnimatedVisibility
import androidx.compose.animation.animateContentSize
import androidx.compose.animation.core.Spring
import androidx.compose.animation.core.animateDpAsState
import androidx.compose.animation.core.animateFloatAsState
import androidx.compose.animation.core.spring
import androidx.compose.animation.core.tween
import androidx.compose.foundation.clickable
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.FlowRow
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.Spacer
import androidx.compose.foundation.layout.defaultMinSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.height
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.size
import androidx.compose.foundation.text.BasicText
import androidx.compose.foundation.text.TextAutoSize
import androidx.compose.foundation.text.selection.SelectionContainer
import androidx.compose.material3.CircularProgressIndicator
import androidx.compose.material3.LocalContentColor
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.ProgressIndicatorDefaults
import androidx.compose.material3.Text
import androidx.compose.material3.ripple
import androidx.compose.runtime.Composable
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.rememberCoroutineScope
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.platform.LocalClipboard
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.platform.toClipEntry
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.semantics.contentDescription
import androidx.compose.ui.semantics.semantics
import androidx.compose.ui.unit.dp
import androidx.compose.ui.viewinterop.AndroidView
import com.byagowi.persiancalendar.R
import com.byagowi.persiancalendar.entities.Calendar
import com.byagowi.persiancalendar.entities.EventsStore
import com.byagowi.persiancalendar.entities.Jdn
import com.byagowi.persiancalendar.global.isAncientIranEnabled
import com.byagowi.persiancalendar.global.isAstronomicalExtraFeaturesEnabled
import com.byagowi.persiancalendar.global.isForcedIranTimeEnabled
import com.byagowi.persiancalendar.global.isTalkBackEnabled
import com.byagowi.persiancalendar.global.language
import com.byagowi.persiancalendar.global.mainCalendar
import com.byagowi.persiancalendar.global.spacedColon
import com.byagowi.persiancalendar.ui.astronomy.LunarAge
import com.byagowi.persiancalendar.ui.theme.appCrossfadeSpec
import com.byagowi.persiancalendar.ui.utils.ItemWidth
import com.byagowi.persiancalendar.utils.MoonInScorpioState
import com.byagowi.persiancalendar.utils.ancientDayName
import com.byagowi.persiancalendar.utils.calculateDaysDifference
import com.byagowi.persiancalendar.utils.formatDate
import com.byagowi.persiancalendar.utils.formatDateAndTime
import com.byagowi.persiancalendar.utils.formatNumber
import com.byagowi.persiancalendar.utils.generateZodiacInformation
import com.byagowi.persiancalendar.utils.getA11yDaySummary
import com.byagowi.persiancalendar.utils.jalaliName
import com.byagowi.persiancalendar.utils.monthName
import com.byagowi.persiancalendar.utils.moonInScorpioState
import com.byagowi.persiancalendar.utils.persianDayOfYear
import com.byagowi.persiancalendar.utils.toGregorianCalendar
import com.byagowi.persiancalendar.utils.toLinearDate
import io.github.cosinekitty.astronomy.Time
import io.github.cosinekitty.astronomy.eclipticGeoMoon
import io.github.cosinekitty.astronomy.seasons
import io.github.cosinekitty.astronomy.sunPosition
import io.github.persiancalendar.calendar.PersianDate
import kotlinx.coroutines.launch
import java.util.Date
import java.util.GregorianCalendar

@Composable
fun CalendarsOverview(
    jdn: Jdn,
    today: Jdn,
    selectedCalendar: Calendar,
    shownCalendars: List<Calendar>,
    isExpanded: Boolean,
) {
    val context = LocalContext.current
    val isToday = today == jdn
    Column(
        Modifier.semantics {
            if (isTalkBackEnabled) this.contentDescription = getA11yDaySummary(
                context.resources,
                jdn,
                isToday,
                EventsStore.empty(),
                withZodiac = true,
                withOtherCalendars = true,
                withTitle = true
            )
        },
    ) {
        Row(
            verticalAlignment = Alignment.CenterVertically,
            modifier = Modifier.padding(horizontal = 24.dp)
        ) {
            this.AnimatedVisibility(isAstronomicalExtraFeaturesEnabled && isExpanded) {
                AndroidView(
                    factory = ::MoonView,
                    update = { it.jdn = jdn.value.toFloat() },
                    modifier = Modifier
                        .padding(end = 8.dp)
                        .size(20.dp)
                )
            }
            val isForcedIranTimeEnabled by isForcedIranTimeEnabled.collectAsState()
            val language by language.collectAsState()
            AnimatedContent(
                if (isToday && isForcedIranTimeEnabled) language.inParentheses.format(
                    jdn.weekDayName, stringResource(R.string.iran_time)
                ) else jdn.weekDayName,
                transitionSpec = appCrossfadeSpec,
                label = "weekday name",
            ) { SelectionContainer { Text(it, color = MaterialTheme.colorScheme.primary) } }
        }
        Spacer(Modifier.height(8.dp))
        CalendarsFlow(shownCalendars, jdn)
        Spacer(Modifier.height(4.dp))

        val date = jdn on selectedCalendar
        val equinox = remember(selectedCalendar, jdn) {
            if (date !is PersianDate) return@remember null
            if (date.month == 12 && date.dayOfMonth >= 20 || date.month == 1 && date.dayOfMonth == 1)
                equinoxTitle(date, jdn, context).first else null
        }

        this.AnimatedVisibility(visible = equinox != null) { CalendarOverviewText(equinox ?: "") }

        this.AnimatedVisibility(!isToday) {
            CalendarOverviewText(
                listOf(
                    stringResource(R.string.days_distance),
                    spacedColon,
                    calculateDaysDifference(context.resources, jdn, today)
                ).joinToString("")
            )
        }

        val language by language.collectAsState()

        val moonInScorpioState = if (isAstronomicalExtraFeaturesEnabled)
            moonInScorpioState(jdn) else null
        this.AnimatedVisibility(moonInScorpioState != null) {
            CalendarOverviewText(
                if (language.isPersian) when (moonInScorpioState) {
                    MoonInScorpioState.Borji -> "قمر در برج عقرب"
                    MoonInScorpioState.Falaki -> "قمر در صورت فلکی عقرب"
                    MoonInScorpioState.Start -> "قمر وارد برج عقرب می‌شود"
                    MoonInScorpioState.End -> "قمر از صورت فلکی عقرب خارج می‌شود"

                    else -> ""
                } else stringResource(R.string.moon_in_scorpio)
            )
        }

        if (language.isPersian) {
            val persianDate = jdn.toPersianDate()
            val dayOfYear = persianDayOfYear(persianDate, jdn)
            val enableExtra = isAncientIranEnabled || isAstronomicalExtraFeaturesEnabled
            this.AnimatedVisibility(enableExtra && isExpanded) {
                CalendarOverviewText(ancientDayName(dayOfYear))
            }
            this.AnimatedVisibility((enableExtra && isExpanded) || persianDate.year < 1304) {
                CalendarOverviewText(jalaliName(persianDate, dayOfYear) + " جلالی")
            }
        }

        this.AnimatedVisibility(isExpanded && isAstronomicalExtraFeaturesEnabled) {
            CalendarOverviewText(
                generateZodiacInformation(context.resources, jdn, withEmoji = true)
            )
        }

        this.AnimatedVisibility(isExpanded && isAstronomicalExtraFeaturesEnabled) {
            val date = jdn.toGregorianCalendar()
            date[GregorianCalendar.HOUR_OF_DAY] = 12
            val time = Time.fromMillisecondsSince1970(date.time.time)
            val lunarAge = LunarAge.fromDegrees(eclipticGeoMoon(time).lon - sunPosition(time).elon)
            val phase = lunarAge.toPhase()
            CalendarOverviewText(phase.emoji + " " + stringResource(phase.stringRes))
        }

        val startOfYearJdn = Jdn(selectedCalendar, date.year, 1, 1)
        val endOfYearJdn = Jdn(selectedCalendar, date.year + 1, 1, 1) - 1
        val currentWeek = jdn.getWeekOfYear(startOfYearJdn)
        val weeksCount = endOfYearJdn.getWeekOfYear(startOfYearJdn)
        val progresses = remember(jdn, selectedCalendar) {
            val (passedDaysInSeason, totalSeasonDays) = jdn.getPositionInSeason()
            val monthLength = selectedCalendar.getMonthLength(date.year, date.month)
            listOfNotNull(
                Triple(R.string.month, date.dayOfMonth, monthLength),
                Triple(R.string.season, passedDaysInSeason, totalSeasonDays),
                Triple(R.string.year, jdn - startOfYearJdn, endOfYearJdn - startOfYearJdn),
            )
        }

        var firstShow by rememberSaveable { mutableStateOf(true) }
        LaunchedEffect(Unit) { firstShow = false }
        val indicatorStrokeWidth by animateDpAsState(
            if (isExpanded && !firstShow) ProgressIndicatorDefaults.CircularStrokeWidth else 0.dp,
            animationSpec = tween(800),
            label = "stroke width",
        )

        this.AnimatedVisibility(isExpanded) {
            Row(
                horizontalArrangement = Arrangement.Center,
                modifier = Modifier
                    .padding(top = 4.dp)
                    .fillMaxWidth(),
            ) {
                progresses.forEach { (stringId, current, max) ->
                    val title = stringResource(stringId)
                    Column(
                        horizontalAlignment = Alignment.CenterHorizontally,
                        modifier = Modifier
                            .semantics {
                                this.contentDescription = "$title$spacedColon$current / $max"
                            }
                            .padding(all = 8.dp),
                    ) {
                        val progress by animateFloatAsState(
                            current.toFloat() / max,
                            animationSpec = spring(
                                dampingRatio = Spring.DampingRatioMediumBouncy,
                                stiffness = Spring.StiffnessLow,
                            ),
                            label = "progress"
                        )
                        CircularProgressIndicator(
                            progress = { progress },
                            strokeWidth = indicatorStrokeWidth
                        )
                        Spacer(modifier = Modifier.height(4.dp))
                        Text(title, style = MaterialTheme.typography.bodyMedium)
                    }
                }
            }
        }

        this.AnimatedVisibility(isExpanded) {
            CalendarOverviewText(
                stringResource(
                    R.string.start_of_year_diff,
                    formatNumber(jdn - startOfYearJdn + 1),
                    formatNumber(currentWeek),
                    formatNumber(date.month)
                )
            )
        }
        this.AnimatedVisibility(isExpanded) {
            CalendarOverviewText(
                stringResource(
                    R.string.end_of_year_diff,
                    formatNumber(endOfYearJdn - jdn),
                    formatNumber(weeksCount - currentWeek),
                    formatNumber(12 - date.month)
                ),
                noTopPadding = true,
            )
        }

        Spacer(Modifier.height(8.dp))
        ExpandArrow(
            modifier = Modifier
                .size(20.dp)
                .align(Alignment.CenterHorizontally),
            isExpanded = isExpanded,
            tint = MaterialTheme.colorScheme.primary,
        )
    }
}

@Composable
private fun CalendarOverviewText(text: String, noTopPadding: Boolean = false) {
    val contextColor = LocalContentColor.current
    Box(
        modifier = Modifier
            .padding(top = (if (noTopPadding) 0 else 4).dp, start = 24.dp, end = 24.dp)
            .fillMaxWidth(),
        contentAlignment = Alignment.Center
    ) {
        SelectionContainer {
            BasicText(
                text = text,
                color = { contextColor },
                style = MaterialTheme.typography.bodyMedium,
                modifier = Modifier.animateContentSize(),
                maxLines = 1,
                softWrap = false,
                autoSize = TextAutoSize.StepBased(
                    minFontSize = MaterialTheme.typography.labelSmall.fontSize,
                    maxFontSize = MaterialTheme.typography.bodyMedium.fontSize,
                ),
            )
        }
    }
}

fun equinoxTitle(date: PersianDate, jdn: Jdn, context: Context): Pair<String, Long> {
    val gregorianYear = jdn.toCivilDate().year
    val timestamp = seasons(gregorianYear).marchEquinox.toMillisecondsSince1970()
    val equinoxYear = when (mainCalendar) {
        Calendar.SHAMSI -> date.year + if (date.month == 12) 1 else 0
        else -> gregorianYear
    }
    val calendar = Date(timestamp).toGregorianCalendar()
    return context.getString(
        R.string.spring_equinox, formatNumber(equinoxYear), calendar.formatDateAndTime()
    ) to timestamp
}

@Composable
private fun CalendarsFlow(calendarsToShow: List<Calendar>, jdn: Jdn) {
    FlowRow(
        modifier = Modifier.fillMaxWidth(),
        horizontalArrangement = Arrangement.Center,
        verticalArrangement = Arrangement.SpaceEvenly,
    ) {
        calendarsToShow.forEach { calendar ->
            val date = jdn on calendar
            Column(
                modifier = Modifier.defaultMinSize(minWidth = ItemWidth.dp),
                horizontalAlignment = Alignment.CenterHorizontally,
            ) {
                val clipboard = LocalClipboard.current
                val coroutineScope = rememberCoroutineScope()
                Column(
                    horizontalAlignment = Alignment.CenterHorizontally,
                    modifier = Modifier
                        .clickable(
                            interactionSource = null,
                            indication = ripple(bounded = false),
                        ) {
                            val entry = ClipData.newPlainText("date", formatDate(date))
                            coroutineScope.launch { clipboard.setClipEntry(entry.toClipEntry()) }
                        }
                        .semantics { this.contentDescription = formatDate(date) },
                ) {
                    Text(
                        formatNumber(date.dayOfMonth),
                        style = MaterialTheme.typography.displayMedium,
                        modifier = Modifier.animateContentSize(),
                    )
                    Text(date.monthName, modifier = Modifier.animateContentSize())
                }
                SelectionContainer {
                    Text(date.toLinearDate(), modifier = Modifier.animateContentSize())
                }
            }
        }
    }
}
